#version 430
#extension GL_ARB_separate_shader_objects : enable










struct SComputeInput
{
	uvec3 GlobalThreadID; uvec3 LocalThreadID; uvec3 GroupID;};

layout( std140) uniform USelectionInfo
{
	uvec4 PositionsOffset;
	uvec4 RadiusOffset;
	uvec4 EnabledOffset;
	vec4 RayOrigin;
	vec4 RayDirection;
	mat4 Planes;
	uint Mode;
} SelectionInfo;
layout(std430 ) readonly buffer BStream
{
	uint Stream[];
} Stream;
layout(std430 ) readonly buffer BStreamInfo
{
	uint StreamInfo[];
} StreamInfo;
layout(std430 ) buffer BSelection
{
	uint Selection[];
} Selection;



bool _RayTraceSphere(vec3 sphereCenter, float sphereRadius, bool enabled  )
{
	if (!enabled)
		return false;

	const vec3 offset = sphereCenter -  SelectionInfo.RayOrigin .xyz;
	const float dist = dot(offset,  SelectionInfo.RayDirection .xyz);
	if (dist <= 0.0f)
		return false;

	const float off2 = dot(offset, offset);
	const float r2 = sphereRadius * sphereRadius;
	if (off2 <= r2)
		return true;

	const float tRel = r2 - (off2 - dist * dist);
	if (tRel < 0.0f)
		return false;

	return (dist - sqrt(tRel) >= 0.f);
}

bool _PointInside4Planes(vec3 coord, bool enabled  )
{
	if (!enabled)
		return false;

	const vec4 dots =  SelectionInfo.Planes [0]  * coord.x +
						SelectionInfo.Planes [1]  * coord.y +
						SelectionInfo.Planes [2]  * coord.z +
						SelectionInfo.Planes [3] ;

	return (dots.x >= 0.f && dots.y >= 0.f && dots.z >= 0.f && dots.w >= 0.f);
}

void ComputeMain( in SComputeInput  cInput  )
{
	const uint particleID = cInput.GlobalThreadID.x;
	const uint count =  StreamInfo.StreamInfo [0] ;
	if (particleID >= count)
		return;

	const float DebugColorNo = 0.f;
	const float DebugColorWhite = 2.0f;
	if ( SelectionInfo.Mode  == 2)
	{
		Selection.Selection [ particleID ] = floatBitsToUint(DebugColorWhite) ; ;
		return;
	}
	const bool modeRay =  SelectionInfo.Mode  == 0;

	bool hasHit = false;
	const uint pOffsetx =  SelectionInfo.PositionsOffset .x;
	if (pOffsetx != uint(-1))
	{
		const uint eOffset =  SelectionInfo.EnabledOffset .x;
		const vec3 sphereCenter =  vec3( uintBitsToFloat(Stream.Stream [pOffsetx + 3 * particleID ]) , uintBitsToFloat(Stream.Stream [(pOffsetx + 3 * particleID ) + 1]) , uintBitsToFloat(Stream.Stream [(pOffsetx + 3 * particleID ) + 2]) ) ;
		const uint enabled = eOffset != uint(-1) ?  Stream.Stream [eOffset + particleID ]  : 1u;
		if (modeRay)
		{
			const uint rOffset =  SelectionInfo.RadiusOffset .x;
			const float sphereRadius = rOffset != uint(-1) ?  uintBitsToFloat(Stream.Stream [rOffset + particleID ])  : 0.f;
			hasHit = _RayTraceSphere(sphereCenter, sphereRadius, (enabled != 0)  );
		}
		else
		{
			hasHit = _PointInside4Planes(sphereCenter, (enabled != 0)  );
		}
	}
	const uint pOffsety =  SelectionInfo.PositionsOffset .y;
	if (pOffsety != uint(-1) && !hasHit)
	{
		const uint eOffset =  SelectionInfo.EnabledOffset .y;
		const vec3 sphereCenter =  vec3( uintBitsToFloat(Stream.Stream [pOffsety + 3 * particleID ]) , uintBitsToFloat(Stream.Stream [(pOffsety + 3 * particleID ) + 1]) , uintBitsToFloat(Stream.Stream [(pOffsety + 3 * particleID ) + 2]) ) ;
		const uint enabled = eOffset != uint(-1) ?  Stream.Stream [eOffset + particleID ]  : 1u;
		if (modeRay)
		{
			const uint rOffset =  SelectionInfo.RadiusOffset .y;
			const float sphereRadius = rOffset != uint(-1) ?  uintBitsToFloat(Stream.Stream [rOffset + particleID ])  : 0.f;
			hasHit = _RayTraceSphere(sphereCenter, sphereRadius, (enabled != 0)  );
		}
		else
		{
			hasHit = _PointInside4Planes(sphereCenter, (enabled != 0)  );
		}
	}
	const uint pOffsetz =  SelectionInfo.PositionsOffset .z;
	if (pOffsetz != uint(-1) && !hasHit)
	{
		const uint eOffset =  SelectionInfo.EnabledOffset .z;
		const vec3 sphereCenter =  vec3( uintBitsToFloat(Stream.Stream [pOffsetz + 3 * particleID ]) , uintBitsToFloat(Stream.Stream [(pOffsetz + 3 * particleID ) + 1]) , uintBitsToFloat(Stream.Stream [(pOffsetz + 3 * particleID ) + 2]) ) ;
		const uint enabled = eOffset != uint(-1) ?  Stream.Stream [eOffset + particleID ]  : 1u;
		if (modeRay)
		{
			const uint rOffset =  SelectionInfo.RadiusOffset .z;
			const float sphereRadius = rOffset != uint(-1) ?  uintBitsToFloat(Stream.Stream [rOffset + particleID ])  : 0.f;
			hasHit = _RayTraceSphere(sphereCenter, sphereRadius, (enabled != 0)  );
		}
		else
		{
			hasHit = _PointInside4Planes(sphereCenter, (enabled != 0)  );
		}
	}
	const uint pOffsetw =  SelectionInfo.PositionsOffset .w;
	if (pOffsetw != uint(-1) && !hasHit)
	{
		const uint eOffset =  SelectionInfo.EnabledOffset .w;
		const vec3 sphereCenter =  vec3( uintBitsToFloat(Stream.Stream [pOffsetw + 3 * particleID ]) , uintBitsToFloat(Stream.Stream [(pOffsetw + 3 * particleID ) + 1]) , uintBitsToFloat(Stream.Stream [(pOffsetw + 3 * particleID ) + 2]) ) ;
		const uint enabled = eOffset != uint(-1) ?  Stream.Stream [eOffset + particleID ]  : 1u;
		if (modeRay)
		{
			const uint rOffset =  SelectionInfo.RadiusOffset .w;
			const float sphereRadius = rOffset != uint(-1) ?  uintBitsToFloat(Stream.Stream [rOffset + particleID ])  : 0.f;
			hasHit = _RayTraceSphere(sphereCenter, sphereRadius, (enabled != 0)  );
		}
		else
		{
			hasHit = _PointInside4Planes(sphereCenter, (enabled != 0)  );
		}
	}
	const float selectedValue = hasHit ? DebugColorWhite : DebugColorNo;





	Selection.Selection [ particleID ] = floatBitsToUint(selectedValue) ; ;
}


layout(local_size_x = 128, local_size_y = 1, local_size_z = 1) in;
void main()
{
	SComputeInput cInput;
	cInput.GlobalThreadID = gl_GlobalInvocationID;
	cInput.LocalThreadID = gl_LocalInvocationID;
	cInput.GroupID = gl_WorkGroupID;
	ComputeMain(cInput);
}
